/*    
 *    junction_seq_construction.cpp		
 *    MapSplice
 *
 *    Copyright (C) 2010 University of Kentucky and
 *                       Zeng Zheng
 *
 *    Authors: Zeng Zheng
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string>
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <vector>
#include <algorithm> 
#include <map>
#include "junction_data_struct.h"
#include "junction_file.h"
#include "junction_sequence.h"
#include "sequence_file.h"

const int NUM_BP_PER_LINE=70;

inline bool sort_junc(const Junction& junc1, const Junction& junc2)
{
	if(junc1.chrom!= junc2.chrom)
		return junc1.chrom < junc2.chrom;
	else if(junc1.start!=junc2.start)
		return junc1.start <junc2.start;
	else
		return junc1.end<junc2.end;
}

inline bool sort_junc_seq_by_name(const Junc_Seq& junc1, const Junc_Seq& junc2)
{
	return junc1.junc_name<junc2.junc_name;
}

inline char complement(int i) 
{
	static const int b2c_size = 20;
	static const char b2c[] = {'T','N','G','N','N','N','C','N','N','N','N','N','N','N','N','N','N','N','N','A'};
	static const char b2cl[] = {'t','n','g','n','n','n','c','n','n','n','n','n','n','n','n','n','n','n','n','a'};
	if (i - 'A' >= 0 && i - 'A' < b2c_size)
		return b2c[i - 'A'];
	else if (i - 'a' >= 0 && i - 'a' < b2c_size)
		return b2cl[i - 'a'];
	else return 'N';
}

inline string revcomp(const string& s) 
{
	string r;
	transform(s.begin(), s.end(), back_inserter(r), complement);
	reverse(r.begin(), r.end());
	return r;
}

class Junction_Seq_Construction
{
public:
	int min_anchor;
	int max_anchor;
	int max_seq_thresh;
	int total_seq_num_for_site;
	string normal_jump_code;
	Seq_File sf;
	ofstream output_fs;

	vector<Junction> junction_set;
	map<string, size_t> chrom_start_map;
	map<string, size_t> chrom_end_map;

	vector<Junc_Seq> junction_head_set;
	vector<Junc_Seq> junction_tail_set;
	size_t junc_count;
	size_t junc_seq_count;
	int curr_chrom_len;
	map <string, int> head_seq_ht;
	map <string, int> tail_seq_ht;

	Junction_Seq_Construction(int min_anchor_len, int max_anchor_len, int max_sequence_threshold)
	{
		min_anchor=min_anchor_len;
		max_anchor=max_anchor_len;
		junc_count=0;
	    junc_seq_count=0;
		max_seq_thresh= max_sequence_threshold;
		normal_jump_code.append(int_to_str(max_anchor_len) + "M" );
	}

	void build_chrom_map()
	{
		string curChrom="";
		for(size_t i=0;i<junction_set.size();i++)
		{
			if(curChrom!=junction_set[i].chrom)
			{
				if(curChrom!="")
					chrom_end_map.insert(make_pair(curChrom,(int)i-1));
				chrom_start_map.insert(make_pair(junction_set[i].chrom,(int)i));
				curChrom=junction_set[i].chrom;
			}
		}
		chrom_end_map.insert(make_pair(curChrom,(int)junction_set.size()-1));
	}

	int check_link(int pos1, int pos2, int min_dis, int max_dis)
	{
		if(pos2-pos1 <= min_dis)
			return -1;
		else if(pos2-pos1 >= max_dis -1)
			return 1;
		else
			return 0;
	}

	string int_to_str(int numerical)
	{
		char c[100];
		sprintf(c,"%d",numerical);
		string str(c);
		return str;
	}

	void push_back_seq(Fusion_Type type, Junc_Seq& my_junc_seq)
	{
		if(type == START)
		{
			if(head_seq_ht.find(my_junc_seq.junc_name)==head_seq_ht.end())
			{
				junction_head_set.push_back(my_junc_seq);
				head_seq_ht.insert(make_pair(my_junc_seq.junc_name, 1));
			}
			else if(head_seq_ht[my_junc_seq.junc_name] <= max_seq_thresh)
			{
				if(tail_seq_ht.find(my_junc_seq.junc_name)==tail_seq_ht.end() || head_seq_ht[my_junc_seq.junc_name] * tail_seq_ht[my_junc_seq.junc_name] <= max_seq_thresh)
				{
					junction_head_set.push_back(my_junc_seq);
					head_seq_ht[my_junc_seq.junc_name]++;
				}
			}
		}
		else
		{
			if(tail_seq_ht.find(my_junc_seq.junc_name)==tail_seq_ht.end())
			{
				junction_tail_set.push_back(my_junc_seq);
				tail_seq_ht.insert(make_pair(my_junc_seq.junc_name, 1));
			}
			else if(tail_seq_ht[my_junc_seq.junc_name] <= max_seq_thresh)
			{
				if(head_seq_ht.find(my_junc_seq.junc_name)==head_seq_ht.end() || tail_seq_ht[my_junc_seq.junc_name] * head_seq_ht[my_junc_seq.junc_name] <= max_seq_thresh)
				{
					junction_tail_set.push_back(my_junc_seq);
					tail_seq_ht[my_junc_seq.junc_name]++;
				}
			}
		}
	}

	void generate_junc_head_forward(Junc_Seq& my_junc_seq, size_t curr_junc, size_t min_index, size_t max_index)
	{
		int len_left=max_anchor - (int)my_junc_seq.sequence.length();
		Junc_Seq normal_junc_seq(my_junc_seq);  		                                        // normal path
		normal_junc_seq.start= max(junction_set[curr_junc].start - len_left + 1, 1);
		string start_sequence=sf.get_sequence(normal_junc_seq.start, junction_set[curr_junc].start);
		normal_junc_seq.sequence.insert(0,start_sequence);
		normal_junc_seq.jump_code.insert(0,int_to_str((int)start_sequence.length())+"M");
		push_back_seq(START, normal_junc_seq);
		for(int i=(int)curr_junc - 1;i>=(int)min_index;i--)                                     // jump path
		{
			if(junction_set[i].junc_type==FUSION)
				continue;
			int link_status=check_link(junction_set[i].end, junction_set[curr_junc].start, min_anchor, len_left);
			if(link_status == -1)
				continue;
			else if(link_status == 1)
				break;
			else
			{
				Junc_Seq new_junc_seq(my_junc_seq);
				string inter_sequence=sf.get_sequence(junction_set[i].end, junction_set[curr_junc].start);
				new_junc_seq.sequence.insert(0,inter_sequence);
				new_junc_seq.jump_code.insert(0, int_to_str((int)inter_sequence.length()) + "M" ); 
				new_junc_seq.jump_code.insert(0, int_to_str(junction_set[i].end - junction_set[i].start - 1) + "N" );
				generate_junc_head_forward(new_junc_seq, i, min_index, max_index);
			}
		}	
	}

	void generate_junc_head_backward(Junc_Seq& my_junc_seq, size_t curr_junc, size_t min_index, size_t max_index)
	{
		int len_left=max_anchor - (int)my_junc_seq.sequence.length();
		Junc_Seq normal_junc_seq(my_junc_seq);  		//normal path
		int seq_end_pos=min(normal_junc_seq.end + len_left - 1, curr_chrom_len);
		string start_sequence=sf.get_sequence(normal_junc_seq.end, seq_end_pos);
		normal_junc_seq.sequence.append(start_sequence);
		normal_junc_seq.jump_code.append(int_to_str((int)start_sequence.length())+"M");
		normal_junc_seq.end=seq_end_pos;
		push_back_seq(START, normal_junc_seq);
		for(int i=(int)curr_junc+1;i<=(int)max_index;i++)
		{
			if(junction_set[i].junc_type==FUSION)
				continue;
			int link_status=check_link(my_junc_seq.end, junction_set[i].start, min_anchor, len_left);
			if(link_status == -1)
				continue;
			else if(link_status == 1)
				break;
			else
			{
				Junc_Seq new_junc_seq(my_junc_seq);
				string inter_sequence=sf.get_sequence(new_junc_seq.end , junction_set[i].start);
				new_junc_seq.sequence.append(inter_sequence);
				new_junc_seq.jump_code.append(int_to_str((int)inter_sequence.length()) + "M" ); 
				new_junc_seq.jump_code.append(int_to_str(junction_set[i].end - junction_set[i].start - 1) + "N" );
				new_junc_seq.end+=(int)inter_sequence.length()+junction_set[i].end - junction_set[i].start - 1;
				generate_junc_head_backward(new_junc_seq, i, min_index, max_index);
			}
		}	
	}

	void generate_junc_tail_forward(Junc_Seq& my_junc_seq, size_t curr_junc, size_t min_index, size_t max_index)
	{
		int len_left=max_anchor - (int)my_junc_seq.sequence.length();
		Junc_Seq normal_junc_seq(my_junc_seq);  		//normal path
		int seq_end_pos=min(junction_set[curr_junc].end + len_left - 1, curr_chrom_len);
		string end_sequence=sf.get_sequence(junction_set[curr_junc].end, seq_end_pos);
		normal_junc_seq.sequence.append(end_sequence);
		normal_junc_seq.jump_code.append( int_to_str((int)end_sequence.length())+"M");
		normal_junc_seq.end=seq_end_pos;
		push_back_seq(END, normal_junc_seq);
		for(int i=(int)curr_junc+1;i<=(int)max_index;i++)
		{
			if(junction_set[i].junc_type==FUSION)
				continue;
			int link_status=check_link(junction_set[curr_junc].end, junction_set[i].start, min_anchor, len_left);
			if(link_status == -1)       // too close
				continue;
			else if(link_status == 1)   // too far
				break;
			else                        // fit
			{
				Junc_Seq new_junc_seq(my_junc_seq);
				string inter_sequence=sf.get_sequence(junction_set[curr_junc].end, junction_set[i].start);
				new_junc_seq.sequence.append(inter_sequence);
				new_junc_seq.jump_code.append( int_to_str((int)inter_sequence.length()) + "M" ); 
				new_junc_seq.jump_code.append(int_to_str(junction_set[i].end - junction_set[i].start - 1) + "N" );
				generate_junc_tail_forward(new_junc_seq, i, min_index, max_index);
			}
		}	
	}

	void generate_junc_tail_backward(Junc_Seq& my_junc_seq, size_t curr_junc, size_t min_index, size_t max_index)
	{
		int len_left=max_anchor - (int)my_junc_seq.sequence.length();
		Junc_Seq normal_junc_seq(my_junc_seq);  		//normal path
		int seq_start_pos =  max(normal_junc_seq.start - len_left + 1, 1);
		string end_sequence=sf.get_sequence(seq_start_pos, normal_junc_seq.start);
		normal_junc_seq.sequence.insert(0,end_sequence);
		normal_junc_seq.jump_code.insert(0,int_to_str((int)end_sequence.length())+"M");
		normal_junc_seq.start=seq_start_pos;
		push_back_seq(END, normal_junc_seq);
		for(int i=(int)curr_junc - 1;i>=(int)min_index;i--)
		{
			if(junction_set[i].junc_type==FUSION)
				continue;
			int link_status=check_link(junction_set[i].end, my_junc_seq.start, min_anchor, len_left);
			if(link_status == -1)       // too close
				continue;
			else if(link_status == 1)   // too far
				break;
			else                        // fit
			{
				Junc_Seq new_junc_seq(my_junc_seq);
				string inter_sequence=sf.get_sequence(junction_set[i].end, new_junc_seq.start);
				new_junc_seq.sequence.insert(0, inter_sequence);
				new_junc_seq.jump_code.insert(0, int_to_str((int)inter_sequence.length()) + "M" ); 
				new_junc_seq.jump_code.insert(0, int_to_str(junction_set[i].end - junction_set[i].start - 1) + "N" );
				new_junc_seq.start-=(int)inter_sequence.length()+junction_set[i].end - junction_set[i].start - 1;
				generate_junc_tail_backward(new_junc_seq, i, min_index, max_index);
			}
		}	
	}

	void output_junc_seq2()
	{
		size_t tail_start_index=0;
		size_t tail_end_index=0;
		string current_junc_name="";
		int total_num=0;
		for(size_t i=0;i<junction_head_set.size();i++)
		{
			if(junction_head_set[i].junc_name!=current_junc_name)	
			{
				cout<<junction_head_set[i].junc_name<<"\t"<<total_num;
				total_num=0;
				tail_start_index=tail_end_index;
				current_junc_name=junction_head_set[i].junc_name;
			}
			for(size_t j=tail_start_index;j<junction_tail_set.size();j++)
			{
				if(junction_head_set[i].junc_name==junction_tail_set[j].junc_name)
				{
					total_num++;
					output_fs<<">"<<junction_head_set[i].chrom<<"_"<<junction_head_set[i].strand<<"_"<<junction_head_set[i].start<<":"<<junction_head_set[i].jump_code<<";"<<junction_tail_set[j].chrom<<"_"<<junction_tail_set[j].strand<<"_"<<junction_tail_set[j].start<<":"<<junction_tail_set[j].jump_code<<endl;	
					string whole_sequence;
					if(junction_head_set[i].strand=="+" && junction_tail_set[j].strand=="+")
					{
						whole_sequence=junction_head_set[i].sequence + junction_tail_set[j].sequence;
					}
					else if(junction_head_set[i].strand=="+" && junction_tail_set[j].strand=="-")
					{
						whole_sequence=junction_head_set[i].sequence + revcomp(junction_tail_set[j].sequence);
					}
					else if(junction_head_set[i].strand=="-" && junction_tail_set[j].strand=="+")
					{
						whole_sequence=revcomp(junction_head_set[i].sequence) + junction_tail_set[j].sequence;
					}
					else
					{
						whole_sequence=revcomp(junction_head_set[i].sequence) + revcomp(junction_tail_set[j].sequence);
					}
					size_t start = 0;
					size_t end= start + NUM_BP_PER_LINE;
					while(true)
					{
						if(end<=whole_sequence.length())
						{
							output_fs<<whole_sequence.substr(start,NUM_BP_PER_LINE)<<endl;
							if(end==whole_sequence.length())
								break;
						}
						else
						{
							output_fs<<whole_sequence.substr(start,whole_sequence.length() - start)<<endl;
							break;
						}
						start=end;
						end=start + NUM_BP_PER_LINE;
					}
				}
				else
				{
					tail_end_index=j;
					break;
				}
			}
		}
	}

	void output_junc_seq()
	{
		cout<<"total head seq: "<<junction_head_set.size()<<endl;
		cout<<"total tail seq: "<<junction_tail_set.size()<<endl;
		size_t head_start_index=0;
		size_t head_end_index=0;
		size_t tail_start_index=0;
		size_t tail_end_index=0;
		while(head_start_index < junction_head_set.size() && tail_start_index < junction_tail_set.size())
		{
			size_t i=head_start_index + 1;
			for(; i<junction_head_set.size();i++)
			{
				if(junction_head_set[i].junc_name != junction_head_set[head_start_index].junc_name)
					break;
			}
			head_end_index=i-1;
			size_t j=tail_start_index + 1;
			for(; j<junction_tail_set.size();j++)
			{
				if(junction_tail_set[j].junc_name != junction_tail_set[tail_start_index].junc_name)
					break;
			}
			tail_end_index=j-1;
			bool too_many_seq=((int)head_end_index - (int)head_start_index + 1) * ((int)tail_end_index - (int)tail_start_index +1) > max_seq_thresh;
			for(i = head_start_index;i<= head_end_index;i++)
			{
				for(j=tail_start_index;j<=tail_end_index;j++)
				{
					if(too_many_seq && (junction_head_set[i].jump_code != normal_jump_code || junction_tail_set[j].jump_code != normal_jump_code))
						continue;
					output_fs<</*junction_head_set[i].junc_name<<*/">"<<junction_head_set[i].chrom<<"_"<<junction_head_set[i].strand<<"_"<<junction_head_set[i].start<<":"<<junction_head_set[i].jump_code<<";"<<junction_tail_set[j].chrom<<"_"<<junction_tail_set[j].strand<<"_"<<junction_tail_set[j].start<<":"<<junction_tail_set[j].jump_code<<endl;	
					string whole_sequence;
					if(junction_head_set[i].strand=="+" && junction_tail_set[j].strand=="+")
					{
						whole_sequence=junction_head_set[i].sequence + junction_tail_set[j].sequence;
					}
					else if(junction_head_set[i].strand=="+" && junction_tail_set[j].strand=="-")
					{
						whole_sequence=junction_head_set[i].sequence + revcomp(junction_tail_set[j].sequence);
					}
					else if(junction_head_set[i].strand=="-" && junction_tail_set[j].strand=="+")
					{
						whole_sequence=revcomp(junction_head_set[i].sequence) + junction_tail_set[j].sequence;
					}
					else
					{
						whole_sequence=revcomp(junction_head_set[i].sequence) + revcomp(junction_tail_set[j].sequence);
					}
					size_t start = 0;
					size_t end= start + NUM_BP_PER_LINE;
					while(true)
					{
						if(end<=whole_sequence.length())
						{
							output_fs<<whole_sequence.substr(start,NUM_BP_PER_LINE)<<endl;
							if(end==whole_sequence.length())
								break;
						}
						else
						{
							output_fs<<whole_sequence.substr(start,whole_sequence.length() - start)<<endl;
							break;
						}
						start=end;
						end=start + NUM_BP_PER_LINE;
					}
				}
			}
			head_start_index=head_end_index+1;
			tail_start_index=tail_end_index+1;
		}
	}

	void construct_junc_seq(char* junc_file_normal, char* junc_file_fusion, char* refseq_path, char* output_file)
	{
		output_fs.open(output_file);
		if( !output_fs ) 
		{
			fprintf(stderr,"error: write junction sequence file error\n");exit(1);
		} 		
		Junction_File junc_normal(junc_file_normal);
		junc_normal.load_all_junctions(junction_set); //load all junctions
		junc_normal.close();

		Junction_File junc_fusion(junc_file_fusion);
		junc_fusion.load_all_fusion_junctions(junction_set); //load all junctions
		junc_fusion.close();
		sort(junction_set.begin(),junction_set.end(),sort_junc); //search support for junc start
		build_chrom_map();

		map<string, size_t>::iterator it_start = chrom_start_map.begin();
		map<string, size_t>::iterator it_end = chrom_end_map.begin();
		while(it_start!=chrom_start_map.end()&&it_end!=chrom_end_map.end())
		{
			sf.init(refseq_path + it_start->first + ".fa");
			sf.load_next_chrom_seq();
			curr_chrom_len=sf.get_seq_len();
			for(size_t i=it_start->second;i<=it_end->second;i++)
			{
				if(junction_set[i].junc_type==FUSION)
				{
					if(junction_set[i].fusion_type==START)
					{
						Junc_Seq head_junc_seq;   
						head_junc_seq.junc_name=junction_set[i].junc_name;
						head_junc_seq.strand=junction_set[i].strand;
						head_junc_seq.chrom=junction_set[i].chrom;
						if(junction_set[i].strand=="+")
						{
							head_junc_seq.end = junction_set[i].start;
							generate_junc_head_forward(head_junc_seq, i,it_start->second, it_end->second);
						}
						else
						{
							head_junc_seq.start = junction_set[i].start;
							head_junc_seq.end = junction_set[i].start;
							generate_junc_head_backward(head_junc_seq, i,it_start->second, it_end->second);
						}
					}
					else  //end
					{
						Junc_Seq tail_junc_seq;   
						tail_junc_seq.junc_name=junction_set[i].junc_name;
						tail_junc_seq.strand=junction_set[i].strand;
						tail_junc_seq.chrom=junction_set[i].chrom;
						if(junction_set[i].strand=="+")
						{
							tail_junc_seq.start=junction_set[i].end;
							generate_junc_tail_forward(tail_junc_seq, i,it_start->second, it_end->second);
						}
						else
						{
							tail_junc_seq.start=junction_set[i].end;
							tail_junc_seq.end=junction_set[i].end;
							generate_junc_tail_backward(tail_junc_seq, i,it_start->second, it_end->second);
						}
					}
				}
			}
			it_start++;
			it_end++;
		}
		sf.close();
		sort(junction_head_set.begin(),junction_head_set.end(), sort_junc_seq_by_name);
		sort(junction_tail_set.begin(),junction_tail_set.end(), sort_junc_seq_by_name);
		output_junc_seq();
		output_fs.close();

	}
};

void print_usage()
{
	fprintf(stderr,"junc_db [min_anchor] [max_anchor] [max_sequence_threshold] [normal_junction_file] [fusion_junction_file] [sequence_path] [output_file]\n");
	exit(1);
}

int main(int argc, char** argv)
{
	if(argc ==1 )
		print_usage();
	else if (argc < 8)
	{
		fprintf(stderr,"error: too few arguments\n");
		print_usage();
		exit(1);
	} 
	else if (argc > 8)
	{
		fprintf(stderr,"error: too many arguments\n");
		print_usage();
		exit(1);
	}

	Junction_Seq_Construction my_construct(atoi(argv[1]), atoi(argv[2]), atoi(argv[3]));
	my_construct.construct_junc_seq(argv[4],argv[5],argv[6],argv[7]);

	//Junction_Seq_Construction my_construct(20, 99, 50);
	//my_construct.construct_junc_seq("D://fusion_db//normal_junc", "D://fusion_db//fusion_remap_junction.unique", "D://fusion_db//refgenome//","D://fusion_db//test.fa");
}
